#include "gen_amd64_regalloc_if.h"


ir_op *op_amd64_Push = NULL;
ir_op *op_amd64_Add = NULL;
ir_op *op_amd64_Immediate = NULL;

ir_op *get_op_amd64_Push(void)         { return op_amd64_Push; }
int    is_amd64_Push(const ir_node *n) { return get_amd64_irn_opcode(n) == iro_amd64_Push; }

ir_op *get_op_amd64_Add(void)         { return op_amd64_Add; }
int    is_amd64_Add(const ir_node *n) { return get_amd64_irn_opcode(n) == iro_amd64_Add; }

ir_op *get_op_amd64_Immediate(void)         { return op_amd64_Immediate; }
int    is_amd64_Immediate(const ir_node *n) { return get_amd64_irn_opcode(n) == iro_amd64_Immediate; }



static int amd64_opcode_start = -1;
static int amd64_opcode_end   = -1;


/** A tag for the amd64 opcodes. Note that the address is used as a tag value, NOT the FOURCC code. */
#define amd64_op_tag FOURCC('A', 'M', 'D', '6')

/** Return the opcode number of the first amd64 opcode. */
int get_amd64_opcode_first(void) {
	return amd64_opcode_start;
}

/** Return the opcode number of the last amd64 opcode + 1. */
int get_amd64_opcode_last(void) {
	return amd64_opcode_end;
}

/** Return 1 if the given opcode is a amd64 machine op, 0 otherwise */
int is_amd64_op(const ir_op *op) {
	return get_op_tag(op) == amd64_op_tag;
}

/** Return 1 if the given node is a amd64 machine node, 0 otherwise */
int is_amd64_irn(const ir_node *node) {
	return is_amd64_op(get_irn_op(node));
}

int get_amd64_irn_opcode(const ir_node *node) {
	if (is_amd64_irn(node))
		return get_irn_opcode(node) - amd64_opcode_start;
	return -1;
}

#ifdef BIT
#undef BIT
#endif
#define BIT(x)  (1 << (x % 32))

static const unsigned amd64_limit_gp_rsp[] = { BIT(REG_RSP), 0 };

static const arch_register_req_t amd64_requirements_gp_gp = {
	arch_register_req_type_normal,
	& amd64_reg_classes[CLASS_amd64_gp],
	NULL,        /* limit bitset */
	0,           /* same pos */
	0            /* different pos */
};


static const arch_register_req_t amd64_requirements__none = {
	arch_register_req_type_none,
	NULL,                         /* regclass */
	NULL,                         /* limit bitset */
	0,                            /* same pos */
	0                             /* different pos */
};


static const arch_register_req_t amd64_requirements_gp_rsp_I_S = {
	arch_register_req_type_ignore | arch_register_req_type_produces_sp | arch_register_req_type_limited,
	& amd64_reg_classes[CLASS_amd64_gp],
	amd64_limit_gp_rsp,
	0,        /* same pos */
	0        /* different pos */
};


static const arch_register_req_t amd64_requirements_gp_rsp = {
	arch_register_req_type_limited,
	& amd64_reg_classes[CLASS_amd64_gp],
	amd64_limit_gp_rsp,
	0,        /* same pos */
	0        /* different pos */
};


static const arch_register_req_t amd64_requirements_gp_in_r1 = {
	arch_register_req_type_should_be_same,
	& amd64_reg_classes[CLASS_amd64_gp],
	NULL,
	1,        /* same pos */
	0        /* different pos */
};



/**
 * construct Push node
 */
ir_node *new_bd_amd64_Push(dbg_info *dbgi, ir_node *block, ir_node *base, ir_node *index, ir_node *mem, ir_node *val, ir_node *stack)
{
	ir_node        *res;
	ir_op          *op      = op_amd64_Push;
	int             flags   = 0;
	backend_info_t *info;
	int             arity   = 5;
	ir_node        *in[5];
	int             n_res   = 2;
	ir_mode        *mode    = mode_T;
	static const be_execution_unit_t ***exec_units = NULL;
	static const arch_register_req_t *in_reqs[] =
	{
		& amd64_requirements_gp_gp,
		& amd64_requirements_gp_gp,
		& amd64_requirements__none,
		& amd64_requirements_gp_gp,
		& amd64_requirements_gp_rsp,
	};

	/* construct in array */
	in[0] = base;
	in[1] = index;
	in[2] = mem;
	in[3] = val;
	in[4] = stack;

	/* create node */
	assert(op != NULL);
	res = new_ir_node(dbgi, current_ir_graph, block, op, mode, arity, in);

	/* init node attributes */
		init_amd64_attributes(res, flags, in_reqs, exec_units, n_res);
	
	info = be_get_info(res);
	(void) info; /* avoid potential warning */
	info->out_infos[0].req = &amd64_requirements_gp_rsp_I_S;
info->out_infos[1].req = &amd64_requirements__none;


	/* optimize node */
	res = optimize_node(res);
	irn_vrfy_irg(res, current_ir_graph);

	return res;
}

/**
 * construct Add node
 */
ir_node *new_bd_amd64_Add(dbg_info *dbgi, ir_node *block, ir_node *op0, ir_node *op1)
{
	ir_node        *res;
	ir_op          *op      = op_amd64_Add;
	int             flags   = 0;
	backend_info_t *info;
	int             arity   = 2;
	ir_node        *in[2];
	int             n_res   = 1;
	ir_mode        *mode    = mode_Iu;
	static const be_execution_unit_t ***exec_units = NULL;
	static const arch_register_req_t *in_reqs[] =
	{
		& amd64_requirements_gp_gp,
		& amd64_requirements_gp_gp,
	};

	/* construct in array */
	in[0] = op0;
	in[1] = op1;

	/* flags */
	flags |= arch_irn_flags_rematerializable;

	/* create node */
	assert(op != NULL);
	res = new_ir_node(dbgi, current_ir_graph, block, op, mode, arity, in);

	/* init node attributes */
		init_amd64_attributes(res, flags, in_reqs, exec_units, n_res);
	
	info = be_get_info(res);
	(void) info; /* avoid potential warning */
	info->out_infos[0].req = &amd64_requirements_gp_in_r1;


	/* optimize node */
	res = optimize_node(res);
	irn_vrfy_irg(res, current_ir_graph);

	return res;
}

/**
 * construct Immediate node
 */
ir_node *new_bd_amd64_Immediate(dbg_info *dbgi, ir_node *block, unsigned imm_value)
{
	ir_node        *res;
	ir_op          *op      = op_amd64_Immediate;
	int             flags   = 0;
	backend_info_t *info;
	int             arity   = 0;
	ir_node       **in      = NULL;
	int             n_res   = 1;
	ir_mode        *mode    = mode_Iu;
	static const be_execution_unit_t ***exec_units = NULL;
	static const arch_register_req_t **in_reqs = NULL;

	/* create node */
	assert(op != NULL);
	res = new_ir_node(dbgi, current_ir_graph, block, op, mode, arity, in);

	/* init node attributes */
		init_amd64_attributes(res, flags, in_reqs, exec_units, n_res);	init_amd64_immediate_attributes(res, imm_value);
	
	info = be_get_info(res);
	(void) info; /* avoid potential warning */
	info->out_infos[0].req = &amd64_requirements_gp_gp;


	/* optimize node */
	res = optimize_node(res);
	irn_vrfy_irg(res, current_ir_graph);

	return res;
}



/**
 * Creates the amd64 specific Firm machine operations
 * needed for the assembler irgs.
 */
void amd64_create_opcodes(const arch_irn_ops_t *be_ops) {
#define N   irop_flag_none
#define L   irop_flag_labeled
#define C   irop_flag_commutative
#define X   irop_flag_cfopcode
#define I   irop_flag_ip_cfopcode
#define F   irop_flag_fragile
#define Y   irop_flag_forking
#define H   irop_flag_highlevel
#define c   irop_flag_constlike
#define K   irop_flag_keep
#define M   irop_flag_machine
#define O   irop_flag_machine_op
#define NB  irop_flag_dump_noblock
#define NI  irop_flag_dump_noinput
#define n   irop_flag_cse_neutral
#define R   (irop_flag_user << 0)

	ir_op_ops  ops;
	int        cur_opcode;
	static int run_once = 0;

	if (run_once)
		return;
	run_once = 1;

	cur_opcode = get_next_ir_opcodes(iro_amd64_last);

	amd64_opcode_start = cur_opcode;

	memset(&ops, 0, sizeof(ops));
	ops.be_ops        = be_ops;
	ops.dump_node     = amd64_dump_node;
	ops.node_cmp_attr = cmp_amd64_attr;
	op_amd64_Push = new_ir_op(cur_opcode + iro_amd64_Push, "amd64_Push", op_pin_state_exc_pinned, N|M, oparity_any, 0, sizeof(amd64_attr_t), &ops);
	set_op_tag(op_amd64_Push, amd64_op_tag);

	memset(&ops, 0, sizeof(ops));
	ops.be_ops        = be_ops;
	ops.dump_node     = amd64_dump_node;
	ops.node_cmp_attr = cmp_amd64_attr;
	op_amd64_Add = new_ir_op(cur_opcode + iro_amd64_Add, "amd64_Add", op_pin_state_exc_pinned, N|M, oparity_binary, 0, sizeof(amd64_attr_t), &ops);
	set_op_tag(op_amd64_Add, amd64_op_tag);

	memset(&ops, 0, sizeof(ops));
	ops.be_ops        = be_ops;
	ops.dump_node     = amd64_dump_node;
	ops.node_cmp_attr = cmp_amd64_attr_immediate;
	op_amd64_Immediate = new_ir_op(cur_opcode + iro_amd64_Immediate, "amd64_Immediate", op_pin_state_floats, c|M, oparity_zero, 0, sizeof(amd64_immediate_attr_t), &ops);
	set_op_tag(op_amd64_Immediate, amd64_op_tag);

	amd64_opcode_end = cur_opcode + iro_amd64_last;
}
